/*
Reveal RootKit is intended to run out of cron or similar services on a regular
base and avoids verbose output as long as nothing was found.

revealrk searches for hidden processes.  If you have a kernel mod rootkit
loaded into memory without hiding any process, don't expect to find anything.

revealrk is inspired by the fine work of Yago Jesus, see:

    http://sourceforge.net/projects/unhide/

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

*******************************************************************************
*
* written by  jk - Juergen.Kahnert@DESY.de
*
* history:
*
* 2013-01-02  jk  first public release version 1.0
*/

/* needed for sched_getaffinity(), getpgid and getsid */
#define _GNU_SOURCE
#ifdef SCHED_GETAFFINITY2
#define __USE_GNU
#endif

#ifdef CONFIG_H
/* make stuff - not needed for scons */
#include "config.h"
#endif

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <dirent.h>
#include <limits.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <sys/resource.h>
#if defined(HAS_TKILL) || defined(HAS_PTHREAD)
#include <sys/syscall.h>
#endif
#ifdef OS_SOLARIS
#include <fcntl.h>
#include <sys/procfs.h>
#endif
#ifdef HAS_SYS_WAIT_H
#include <sys/wait.h>
#else
#include <wait.h>
#endif
#ifdef HAS_SCHEDULER_H
#include <sched.h>
#endif
#if defined(HAS_PTHREAD) && defined(OS_LINUX)
#include <pthread.h>
#endif
#ifdef HAS_CAPABILITY
#include <sys/capability.h>
#ifdef _POSIX_SOURCE
extern int capgetp(pid_t, cap_t);
#endif
#endif

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#ifdef OS_LINUX
#if defined(USE_OLD_PS)
/* old Linux ps command, not supporting threads */
#define PS_ALL "ps --no-header -e o pid"
#define PS_PGID "ps --no-header -e o pgid | sort -nu"
#define PS_SESS "ps --no-header -e o sess | sort -nu"
#elif defined(USE_OLD_PS_WITH_THREAD)
/* old Linux ps command, supporting threads */
#define PS_ALL "ps --no-header -em o pid"
#define PS_PGID "ps --no-header -em o pgid | sort -nu"
#define PS_SESS "ps --no-header -em o sess | sort -nu"
#else
/* ps command for Linux systems */
#define PS_ALL "ps --no-header -eL o lwp"
#define PS_PGID "ps --no-header -eL o pgid | sort -nu"
#define PS_SESS "ps --no-header -eL o sess | sort -nu"
#endif
#elif defined(OS_OPENBSD)
/* ps command for OpenBSD systems */
#define PS_ALL "ps axmk -o pid=''"
#define PS_PGID "ps axmk -o pgid='' | sort -nu"
#define PS_SESS "ps axmk -o sid='' | sort -nu"
#elif defined(OS_BSD)
/* ps command for BSD systems */
#define PS_ALL "ps axm -o pid=''"
#define PS_PGID "ps axm -o pgid='' | sort -nu"
#define PS_SESS "ps axm -o sid='' | sort -nu"
#elif defined(OS_SOLARIS)
/* ps command for Solaris systems */
#define PS_ALL "ps -eL -o pid=''"
#define PS_PGID "ps -eL -o pgid='' | sort -nu"
#define PS_SESS "ps -eL -o sid='' | sort -nu"
#endif

/* command line option errors */
#define INVALID     0
#define MISSING     1

/* type of hidden process */
#define SELFTEST    1
#define HIDDEN      2
#define BRUTE       3
#define FAKE        4

const char* VERSION = "1.0.1";
const char* MAILADDR = "Juergen.Kahnert@DESY.de";
const char* FOOTER = "License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n" \
                     "This is free software: you are free to change and redistribute it.\n" \
                     "There is NO WARRANTY, to the extent permitted by law.\n\n" \
                     "Written by Jürgen Kahnert <Juergen.Kahnert@DESY.de>.\n";
const char* PATH_PID_MAX = "/proc/sys/kernel/pid_max";

/* maximum retries before ending double checking */
int g_nMaxChecks = 64;

/* timeout and exit after how much seconds with no response between checks */
#define DEFAULT_TIMEOUT 300
unsigned g_uCycleTimeoutSecs = DEFAULT_TIMEOUT;

/* disabled timeout for the entire run */
unsigned g_uGlobalTimeoutSecs = 0;

struct pid_list_t
{
    pid_t uPID;
    int nType;
    struct pid_list_t* psoNext;
};

struct kernel_version_t
{
    char* pszName;
    int nRelease;
    int nMajor;
    int nMinor;
    int bHasProcTask;
    int bHasRandomPIDs;
    struct utsname soRaw;
};

struct kernel_version_t g_soKernel;

/* sysctl kernel.pid_max */
pid_t g_uMaxPID = 32768;

/* kernel pids, unable to get via fork */
unsigned g_uReservedPIDs = 300;

/* pointer to program name */
char* g_szProgramName = NULL;

/* callers user id - should be 0 for root */
uid_t g_uUID;

#ifdef OS_LINUX
/* system boot time */
unsigned long g_ulBootTime = 0;
#endif

/* pids stored in bit array */
char* g_payPS_PIDs  = NULL;
char* g_payProcPIDs = NULL;

/* final brute force test to reduce false positives */
int g_nMoreBruteTesting = 0;

/* command line options */
FILE* opt_fsLogFile             = NULL;
char* opt_szMaxPID              = NULL;
int   opt_nDaemon               = 0;
int   opt_nFakeHidden           = 0;
#if defined(OS_LINUX) || defined(OS_SOLARIS)
int   opt_nFakeSearch           = 0;
#endif
int   opt_nForceBrute           = 0;
#ifdef OS_LINUX
int   opt_nIgnoreZombie         = 0;
#endif
int   opt_nListTests            = 0;
int   opt_nNiceness             = 0;
int   opt_nOmitSignature        = 0;
int   opt_nQuiet                = 0;
int   opt_nRandom               = 0;
int   opt_nSyslog               = 0;
int   opt_nVerbose              = 0;
/* tests - system functions */
int   opt_getpgid               = 1;
int   opt_getpriority           = 1;
int   opt_getsid                = 1;
int   opt_kill                  = 1;
#ifdef HAS_SIGQUEUE
int   opt_sigqueue              = 1;
#endif
#ifdef HAS_TKILL
int   opt_tkill                 = 1;
#endif
#ifdef HAS_SCHEDULER_H
int   opt_sched_getparam        = 1;
int   opt_sched_rr_get_interval = 1;
int   opt_sched_getscheduler    = 1;
#ifdef SCHED_GETAFFINITY
int   opt_sched_getaffinity     = 1;
#endif
#endif
#ifdef HAS_CAPABILITY
int   opt_capgetp               = 1;
#endif
/* tests - proc */
int   opt_chdir                 = 1;
int   opt_opendir               = 1;
int   opt_proc                  = 1;
#ifdef OS_LINUX
int   opt_tgid                  = 1;
#endif
int   opt_readdir               = 1;
/* tests - brute force pids */
int   opt_vfork                 = 1;
#if defined(HAS_PTHREAD) && defined(OS_LINUX)
int   opt_pthread               = 1;
#endif
/* other tests */
#if defined(OS_LINUX)
int   opt_fake_name             = 1;
#elif defined(OS_SOLARIS)
int   opt_fake_name             = 0;
#endif


/* current pid test results */
unsigned g_uVisible = 0;
unsigned g_uHidden  = 0;

#if defined(OS_LINUX) && !defined(USE_OLD_PS)
/* status vars for GetTGID function */
pid_t g_gettgid_TGID = 0;
pid_t g_gettgid_PID  = 0;
#endif

void CalcSignature(pid_t);
void SetupCycleWatchdog(void);
int SearchBruteForce(struct pid_list_t**);

void Output(int tnLevel, const char* tcszFormat, ...)
{
    char acBuf[2048];
    va_list pvArgs;

    /* skip output if verbose level isn't high enough */
    if (opt_nVerbose < tnLevel) return;

    va_start(pvArgs, tcszFormat);
    vsnprintf(acBuf, sizeof(acBuf), tcszFormat, pvArgs);
    va_end(pvArgs);
    acBuf[sizeof(acBuf)-1] = '\0';
    if (opt_fsLogFile) fputs(acBuf, opt_fsLogFile);
    else if (!opt_nQuiet) fputs(acBuf, stdout);
    if (opt_nSyslog && tnLevel == 0)
    {
        /* strip newlines from end of message */
        char* pc = acBuf + strlen(acBuf) - 1;
        while (*pc == '\n') *pc-- = '\0';
        pc = acBuf;
        /* strip newlines from start of message */
        while (*pc == '\n') ++pc;
        syslog(LOG_WARNING, "%s", pc);
    }
}

char* BitmapInit(void)
{
    char* ayBitmap = (char*)calloc(1, (g_uMaxPID >> 3) + 1);

    if (!ayBitmap)
    {
        Output(0, "Can't allocate enough memory: %s\n", strerror(errno));
        exit(103);
    }
    return ayBitmap;
}

int BitmapGet(const char* tcayBitmap, pid_t tnPID)
{
    if (!tcayBitmap)
    {
        Output(0, "Can't access empty bitmap!\n");
        exit(104);
    }
    return tcayBitmap[tnPID >> 3] & (1 << (tnPID & 7));
}

void BitmapSet(char* tayBitmap, pid_t tnPID)
{
    if (!tayBitmap)
    {
        Output(0, "Can't access empty bitmap!\n");
        exit(104);
    }
    if (tnPID > g_uMaxPID)
    {
        Output(0, "Error: Found pid %u, but maxpid is %u!\n", tnPID, g_uMaxPID);
        exit(107);
    }
    tayBitmap[tnPID >> 3] |= (1 << (tnPID & 7));
}

void CycleWatchdog(int tnSignal)
{
    Output(0, "%s timed out between checks after %d seconds -> forced exit!\n", g_szProgramName, g_uCycleTimeoutSecs);
    exit(108);
}

void GlobalWatchdog(int tnSignal)
{
    Output(0, "%s timed out entire run after %d seconds -> forced exit!\n", g_szProgramName, g_uGlobalTimeoutSecs);
    exit(108);
}

void GetCmdLine(pid_t tuPID, char* tacBuffer, unsigned tuBufSize)
{
    #ifdef OS_LINUX
    struct stat soStatBuf;
    FILE* fsFile;
    int c, i;
    char* pc;

    snprintf(tacBuffer, tuBufSize, "/proc/%u/cmdline", tuPID);
    if (stat(tacBuffer, &soStatBuf) == 0)
    {
        if ((fsFile = fopen(tacBuffer, "r")))
        {
            i = 0;
            pc = tacBuffer;
            while (++i < tuBufSize && (c = fgetc(fsFile)) != EOF)
            {
                *pc++ = (c == 0) ? ' ' : (char)c;
            }
            if (pc > tacBuffer) --pc;
            *pc = '\0';
            fclose(fsFile);
            return;
        }
    }
    #elif defined(OS_SOLARIS)
    struct prpsinfo soInfo;
    char acBuf[256];
    int nFD;

    snprintf(acBuf, sizeof(acBuf), "/proc/%u", (unsigned)tuPID);
    if ((nFD = open(acBuf, O_RDONLY)) != -1)
    {
        if (ioctl(nFD, PIOCPSINFO, &soInfo) != -1)
        {
            strncpy(tacBuffer, soInfo.pr_psargs, tuBufSize);
            tacBuffer[tuBufSize-1] = '\0';
            close(nFD);
            return;
        }
        close(nFD);
    }
    #endif
    tacBuffer[0] = '\0';
}

char* CheckProcessName(pid_t tuPID, int tnSpaces)
{
    static char acBuf[PATH_MAX];
    char* pc = acBuf;
    int nCount = 0;
    int nMax = 0;

    GetCmdLine(tuPID, acBuf, sizeof(acBuf));
    while (*pc)
    {
        if (*pc == ' ' || *pc == '\t') ++nCount;
        else
        {
            if (nCount > nMax)
            {
                if (nCount >= tnSpaces) return acBuf;
                nMax = nCount;
            }
            nCount = 0;
        }
        ++pc;
    }
    return NULL;
}

#ifdef OS_LINUX
int CheckZombie(pid_t tuPID)
{
    int nRet = 1;
    char acBuf[PATH_MAX];
    FILE* fsProcStat;
    char* pc;

    if (opt_nIgnoreZombie)
    {
        snprintf(acBuf, sizeof(acBuf), "/proc/%u/stat", (unsigned)tuPID);
        if ((fsProcStat = fopen(acBuf, "r")))
        {
            if (fgets(acBuf, sizeof(acBuf), fsProcStat) && (pc = strrchr(acBuf, ')')))
                if (*(pc+2) == 'Z')
                    nRet = 0;
            fclose(fsProcStat);
        }
    }
    return nRet;
}

void GetBootTime(void)
{
    char acBuf[1024];
    FILE* fsStat;

    if ((fsStat = fopen("/proc/stat", "r")))
    {
        /* get system boot time in seconds */
        while (!feof(fsStat))
            if (fgets(acBuf, sizeof(acBuf), fsStat))
                if (!strncmp(acBuf, "btime", 5))
                {
                    sscanf(acBuf+5, "%lu", &g_ulBootTime);
                    break;
                }
        fclose(fsStat);
    }
}
#endif

void GetKernelInfo(void)
{
    char* pszValue;
    char* pc;

    if (uname(&g_soKernel.soRaw) == -1)
    {
        fprintf(stderr, "Error - can't query kernel version: %s\n", strerror(errno));
        exit(101);
    }
    g_soKernel.nMajor = -1;
    g_soKernel.nMinor = -1;
    g_soKernel.bHasProcTask = 0;
    g_soKernel.bHasRandomPIDs = 0;
    g_soKernel.pszName = g_soKernel.soRaw.sysname;
    pc = pszValue = g_soKernel.soRaw.release;
    while (*pc && *pc != '.' && *pc != ' ') ++pc;
    if (!*pszValue)
    {
        fprintf(stderr, "Error - can't identify kernel release: %s\n", pszValue);
        exit(101);
    }
    *pc = '\0';
    sscanf(pszValue, "%d", &g_soKernel.nRelease);
    pszValue = ++pc;
    while (*pc && *pc != '.' && *pc != '-' && *pc != ' ') ++pc;
    if (!*pszValue)
    {
        if (!strcmp(g_soKernel.pszName, "Linux"))
        {
            Output(0, "Can't get linux kernel major release.\n");
            exit(101);
        }
        Output(3, "Debug: Identified %s %d\n", g_soKernel.pszName, g_soKernel.nRelease);
    }
    else
    {
        *pc = '\0';
        sscanf(pszValue, "%d", &g_soKernel.nMajor);
        pszValue = ++pc;
        while (*pc && *pc != '.' && *pc != '-' && *pc != ' ') ++pc;
        *pc = '\0';
        if (*pszValue && isdigit((int)*pszValue))
        {
            sscanf(pszValue, "%d", &g_soKernel.nMinor);
            Output(3, "Debug: Identified %s %d.%d.%d\n", g_soKernel.pszName, g_soKernel.nRelease, g_soKernel.nMajor, g_soKernel.nMinor);
        }
        else Output(3, "Debug: Identified %s %d.%d\n", g_soKernel.pszName, g_soKernel.nRelease, g_soKernel.nMajor);
    }
    if (!strcmp(g_soKernel.pszName, "Linux") && (g_soKernel.nRelease > 2 || (g_soKernel.nRelease == 2 && g_soKernel.nMajor > 4))) g_soKernel.bHasProcTask = 1;
    if (!strcmp(g_soKernel.pszName, "OpenBSD"))
    {
        g_soKernel.bHasRandomPIDs = 1;
        opt_getsid = 0;
        opt_getpgid = 0;
    }
}

void GetMaxPID(void)
{
    FILE* fsMaxPID;
    unsigned uMax;

    if (opt_szMaxPID)
    {
        /* use argument given from command line */
        if (sscanf(opt_szMaxPID, "%u", &uMax) != 1)
        {
            fprintf(stderr, "Error - can't read pid_max out of '%s'\n", opt_szMaxPID);
            exit(101);
        }
        else if (uMax < 1024)
        {
            fprintf(stderr, "Using %u as pid_max seems a bit too low...\n", uMax);
            exit(101);
        }
        else if (uMax < 32768) Output(2, "using pid_max = %u as requested\n", uMax);
        g_uMaxPID = uMax;
    }
    else
    {
        /* auto-detect max pid */
        if (!(fsMaxPID = fopen(PATH_PID_MAX, "r")))
        {
            if (strcmp(g_soKernel.pszName, "Linux") == 0)
            {
                if (g_soKernel.nRelease > 2 || (g_soKernel.nRelease == 2 && g_soKernel.nMajor > 4))
                    Output(2, "Warning - can't get current maximum PID: %s\n", strerror(errno));
                Output(2, "Assuming pid_max is set to %u\n", g_uMaxPID);
            }
            else if (strstr(g_soKernel.pszName, "OpenBSD"))
            {
                g_uMaxPID = 32767;
                g_uReservedPIDs = 1;
                Output(3, "Can't get maximum PID, using %u.\n", g_uMaxPID);
            }
            else if (strstr(g_soKernel.pszName, "BSD"))
            {
                g_uMaxPID = 99999;
                g_uReservedPIDs = 100;
                Output(3, "Can't get maximum PID, using %u.\n", g_uMaxPID);
            }
            else if (strstr(g_soKernel.pszName, "SunOS"))
            {
                g_uMaxPID = 30000;
                g_uReservedPIDs = 1;
                Output(3, "Can't get maximum PID, using %u.\n", g_uMaxPID);
            }
            else Output(2, "Unknown system, using %u as maximum PID.\n", g_uMaxPID);
        }
        else
        {
            if ((fscanf(fsMaxPID, "%u", &uMax) != 1) || uMax < 32768)
            {
                Output(0, "Error - parse file %s, got %u.\n", PATH_PID_MAX, uMax);
            }
            else
            {
                g_uMaxPID = uMax;
            }
            fclose(fsMaxPID);
        }
    }
    Output(3, "Debug: maxpid set to %u.\n", g_uMaxPID);
}

char* GetStateDesc(char tcState)
{
    char* pszDesc = NULL;

    switch (tcState)
    {
        case 'D': pszDesc = "waiting in uninterruptible disk sleep"; break;
        case 'O': pszDesc = "running on a processor"; break;
        case 'R': 
        #ifdef OS_SOLARIS
                  pszDesc = "runnable - process is on a run queue"; break;
        #else
                  pszDesc = "running"; break;
        #endif
        case 'S': pszDesc = "sleeping in an interruptible wait"; break;
        case 'T': pszDesc = "traced or stopped (on a signal)"; break;
        case 'W': pszDesc = "waiting"; break;
        case 'Z': pszDesc = "zombie"; break;
    }
    return pszDesc;
}

#if defined(OS_LINUX) && !defined(USE_OLD_PS)
pid_t GetTGID(pid_t tuPID)
{
    FILE*    fsProcStatus;
    char     acPath[128];
    char     szLine[128];
    char*    pszTGID = szLine + 5;
    int      bFound = 0;
    unsigned uTGID;

    /* speed up if test was called before */
    if (g_gettgid_PID == tuPID) return g_gettgid_TGID;

    snprintf(acPath, sizeof(acPath), "/proc/%u/status", tuPID);
    fsProcStatus = fopen(acPath, "r");
    if (fsProcStatus)
    {
        errno = 0;
        while (!errno && !bFound && !feof(fsProcStatus))
            if (fgets(szLine, sizeof(szLine), fsProcStatus) && !errno)
                if (!strncmp(szLine, "Tgid:", 5))
                    bFound = 1;
        fclose(fsProcStatus);

        if (bFound)
        {
            while (((*pszTGID == ' ') || (*pszTGID == '\t')) && (pszTGID <= szLine+sizeof(szLine)-1)) ++pszTGID;
            sscanf(pszTGID, "%u", &uTGID);
            if (!uTGID && tuPID < 10 && g_soKernel.nMajor <= 4 && g_soKernel.nRelease == 2)
            {
                /* fake tgid for old kernels running in kernel tgid 0 */
                uTGID = tuPID;
            }
            g_gettgid_TGID = uTGID;
            g_gettgid_PID = tuPID;
        }
    }
    return g_gettgid_TGID;
}
#endif

void ListAppend(struct pid_list_t** tppsoList, pid_t tuPID, int tnType)
{
    struct pid_list_t* psoPID = malloc(sizeof(struct pid_list_t));
    struct pid_list_t* psoIter = *tppsoList;

    if (!psoPID)
    {
        Output(0, "Can't allocate enough memory: %s\n", strerror(errno));
        exit(103);
    }
    psoPID->uPID = tuPID;
    psoPID->nType = tnType;
    psoPID->psoNext = NULL;
    if (psoIter != NULL)
    {
        while (psoIter->psoNext != NULL) psoIter = psoIter->psoNext;
        psoIter->psoNext = psoPID;
    }
    else *tppsoList = psoPID;
}

void ListClear(struct pid_list_t* tpsoList)
{
    struct pid_list_t* pso;

    while (tpsoList)
    {
        pso = tpsoList;
        tpsoList = tpsoList->psoNext;
        free(pso);
    }
}

int ListCmp(struct pid_list_t* tpsoList1, struct pid_list_t* tpsoList2)
{
    struct pid_list_t* psoIter1 = tpsoList1;
    struct pid_list_t* psoIter2 = tpsoList2;

    while (psoIter1 && psoIter2)
    {
        if (psoIter1->uPID == psoIter2->uPID)
        {
            psoIter1 = psoIter1->psoNext;
            psoIter2 = psoIter2->psoNext;
        }
        else
        {
            psoIter2 = NULL;
            break;
        }
    }
    return (psoIter1 && !psoIter2) || (!psoIter1 && psoIter2);
}

struct pid_list_t* ListIntersect(struct pid_list_t* tpsoList1, struct pid_list_t* tpsoList2)
{
    struct pid_list_t* psoNewList = NULL;
    struct pid_list_t** ppsoIter = &psoNewList;

    while (tpsoList1 && tpsoList2)
    {
        if (tpsoList1->uPID == tpsoList2->uPID)
        {
            ListAppend(ppsoIter, tpsoList1->uPID, (tpsoList1->nType < tpsoList2->nType) ? tpsoList1->nType : tpsoList2->nType);
            ppsoIter = &((*ppsoIter)->psoNext);
            tpsoList1 = tpsoList1->psoNext;
            tpsoList2 = tpsoList2->psoNext;
        }
        else if (tpsoList1->uPID < tpsoList2->uPID) tpsoList1 = tpsoList1->psoNext;
        else tpsoList2 = tpsoList2->psoNext;
    }
    return psoNewList;
}

unsigned ListLen(struct pid_list_t* tpsoList)
{
    unsigned uLen = 0;
    struct pid_list_t* psoIter = tpsoList;

    while (psoIter)
    {
        ++uLen;
        psoIter = psoIter->psoNext;
    }
    return uLen;
}

int ListRemove(struct pid_list_t** tppsoList, pid_t tuPID)
{
    struct pid_list_t* psoIter = *tppsoList;
    struct pid_list_t* psoPrev = NULL;
    int bDeleted = 0;

    if (*tppsoList && (*tppsoList)->uPID == tuPID)
    {
        if ((*tppsoList)->nType != FAKE)
        {
            /* delete head node of tppsoList */
            *tppsoList = (*tppsoList)->psoNext;
            free(psoIter);
            bDeleted = 1;
        }
    }
    else while (psoIter)
    {
        psoPrev = psoIter;
        psoIter = psoIter->psoNext;
        if (psoIter && psoIter->uPID == tuPID)
        {
            if ((*tppsoList)->nType != FAKE)
            {
                psoPrev->psoNext = psoIter->psoNext;
                free(psoIter);
                bDeleted = 1;
            }
            break;
        }
    }
    return bDeleted;
}

struct pid_list_t* ListRemoveVisible(struct pid_list_t** tppsoList)
{
    char acLine[PATH_MAX];
    unsigned uPID;
    FILE* fsPS;

    /* avoid getcwd errors while inside of /proc/XXX of terminated process */
    if (chdir("/") != 0) Output(3, "Can't chdir /: %s", strerror(errno));
    if (*tppsoList)
    {
        /* session test */
        if (!(fsPS = popen(PS_SESS, "r")))
        {
            Output(0, "Calling '%s' failed: %s\n", PS_SESS, strerror(errno));
            exit(105);
        }
        while (*tppsoList && fgets(acLine, sizeof(acLine), fsPS))
        {
            sscanf(acLine, "%u", &uPID);
            ListRemove(tppsoList, uPID);
        }
    }
    if (*tppsoList)
    {
        /* pgid test */
        if (!(fsPS = popen(PS_PGID, "r")))
        {
            Output(0, "Calling '%s' failed: %s\n", PS_PGID, strerror(errno));
            exit(105);
        }
        while (*tppsoList && fgets(acLine, sizeof(acLine), fsPS))
        {
            sscanf(acLine, "%u", &uPID);
            ListRemove(tppsoList, uPID);
        }
    }
    return *tppsoList;
}

void ListSplitBrute(struct pid_list_t** tppsoList, struct pid_list_t** tppsoBrute, struct pid_list_t** tppsoOther)
{
    struct pid_list_t* psoIter = *tppsoList;

    while (psoIter)
    {
        if (psoIter->nType == BRUTE) ListAppend(tppsoBrute, psoIter->uPID, BRUTE);
        else ListAppend(tppsoOther, psoIter->uPID, psoIter->nType);
        psoIter = psoIter->psoNext;
    }
    ListClear(*tppsoList);
    *tppsoList = NULL;
}

struct pid_list_t* ListUnion(struct pid_list_t* tpsoList1, struct pid_list_t* tpsoList2)
{
    struct pid_list_t* psoNewList = NULL;

    while (tpsoList1 && tpsoList2)
    {
        if (tpsoList1->uPID < tpsoList2->uPID)
        {
            ListAppend(&psoNewList, tpsoList1->uPID, tpsoList1->nType);
            tpsoList1 = tpsoList1->psoNext;
        }
        else if (tpsoList1->uPID == tpsoList2->uPID)
        {
            ListAppend(&psoNewList, tpsoList1->uPID, tpsoList1->nType);
            tpsoList1 = tpsoList1->psoNext;
            tpsoList2 = tpsoList2->psoNext;
        }
        else
        {
            ListAppend(&psoNewList, tpsoList2->uPID, tpsoList2->nType);
            tpsoList2 = tpsoList2->psoNext;
        }
    }
    while (tpsoList1)
    {
        ListAppend(&psoNewList, tpsoList1->uPID, tpsoList1->nType);
        tpsoList1 = tpsoList1->psoNext;
    }
    while (tpsoList2)
    {
        ListAppend(&psoNewList, tpsoList2->uPID, tpsoList2->nType);
        tpsoList2 = tpsoList2->psoNext;
    }
    return psoNewList;
}

void BruteTest(struct pid_list_t** tppsoResult, struct pid_list_t* (*fnBruteTest)(void), const char* tcszName)
{
    struct pid_list_t* psoListTmp = NULL;
    struct pid_list_t* psoListHidden = NULL;
    struct pid_list_t* psoListDoubleCheck = NULL;
    int i;
    int bDone = 0;
    unsigned uHidden;
    unsigned uCount;

    psoListHidden = (*fnBruteTest)();
    uHidden = ListLen(psoListHidden);
    Output(2, "Found %u hidden process%s after step 1 with %s.\n", uHidden, (uHidden > 1) ? "es" : "", tcszName);
    for (i = 2; uHidden && i <= g_nMaxChecks && !bDone; ++i)
    {
        /* give processes some time to exit before starting next round */
        if (g_nMoreBruteTesting) sleep(1);
        psoListDoubleCheck = (*fnBruteTest)();
        if (ListCmp(psoListHidden, psoListDoubleCheck))
        {
            /* lists are not equal */
            psoListTmp = ListIntersect(psoListHidden, psoListDoubleCheck);
            uCount = ListLen(psoListTmp);
            if (uCount == uHidden && !ListCmp(psoListTmp, psoListHidden)) bDone = 1;
            uHidden = uCount;
            ListClear(psoListHidden);
            psoListHidden = psoListTmp;
        }
        else bDone = 1;
        ListClear(psoListDoubleCheck);
        Output(2, "Found %u hidden process%s after step %d with %s.\n", uHidden, (uHidden > 1) ? "es" : "", i, tcszName);
    }
    if (!*tppsoResult) *tppsoResult = psoListHidden;
    else 
    {
        psoListTmp = ListUnion(*tppsoResult, psoListHidden);
        ListClear(*tppsoResult);
        ListClear(psoListHidden);
        *tppsoResult = psoListTmp;
    }
    if (*tppsoResult) ListRemoveVisible(tppsoResult);
}

#ifdef OS_LINUX
void PrintEnviron(pid_t tuPID)
{
    char acBuf[8192];
    FILE* fsEnv;
    char* pc;

    if (!opt_nVerbose) return;
    memset(acBuf, 0, sizeof(acBuf));
    snprintf(acBuf, sizeof(acBuf), "/proc/%u/environ", (unsigned)tuPID);
    if ((fsEnv = fopen(acBuf, "r")))
    {
        Output(1, "  environment variables:\n");
        while (!feof(fsEnv))
        {
            if (fgets(acBuf, sizeof(acBuf), fsEnv))
            {
                pc = acBuf;
                while (pc < acBuf + sizeof(acBuf) && *pc != '\0')
                {
                    Output(1, "  - %s\n", pc);
                    pc += strlen(pc);
                    ++pc;
                }
            }
            else break;
        }
        fclose(fsEnv);
    }
}
#endif

#ifdef OS_LINUX
void PrintStatInfo(pid_t tuPID)
{
    char acTime[128];
    char acBuf[PATH_MAX];
    char* pszDesc;
    char* pc;
    int nField = 3;
    unsigned long ulJiffies = 0;
    time_t lStartTime;
    unsigned uPPID;
    FILE* fsProcStat;

    snprintf(acBuf, sizeof(acBuf), "/proc/%u/stat", (unsigned)tuPID);
    if ((fsProcStat = fopen(acBuf, "r")))
    {
        if (fgets(acBuf, sizeof(acBuf), fsProcStat) && (pc = strrchr(acBuf, ')')))
        {
            /* overwrite ')' to terminate command name */
            *pc = '\0';
            Output(0, "  command: %s\n", strchr(acBuf, '(') + 1);
            /* set pointer to field 3 -> process state */
            pc += 2;
            pszDesc = GetStateDesc(*pc);
            if (pszDesc) Output(0, "  state: %c - %s\n", *pc, pszDesc);
            else Output(0, "  unknown state '%c'\n", *pc);
            /* go to field 4 -> parent pid */
            while (*pc++ != ' ') ; ++nField;
            sscanf(pc, "%u", &uPPID);
            Output(0, "  parent pid: %u\n", uPPID);
            /* find field 22 -> starttime */
            while (nField < 22)
            {
                while (*pc++ != ' ') ;
                ++nField;
            }
            /* get process start time in jiffies */
            sscanf(pc, "%lu", &ulJiffies);
        }
        fclose(fsProcStat);
        if (g_ulBootTime && ulJiffies)
        {
            lStartTime = g_ulBootTime + ulJiffies / 100;
            strftime(acTime, sizeof(acTime), "%Y-%m-%d %H:%M:%S", localtime(&lStartTime));
            Output(0, "  running since: %s\n", acTime);
        }
    }
}
#endif

#ifdef OS_BSD
void PrintStatusInfoBSD(char* tacBuf)
{
    unsigned long ulStartTime;
    time_t lStart;
    char acTime[128];
    char* pcValue = tacBuf;
    char* pc = tacBuf;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  command: %s\n", pcValue);
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  parent pid: %s\n", pcValue);
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  gid: %s\n", pcValue);
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  sid: %s\n", pcValue);
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  term: %s\n", pcValue);
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  flags: %s\n", pcValue);
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue)
    {
        sscanf(pcValue, "%lu", &ulStartTime); lStart = (time_t)ulStartTime;
        strftime(acTime, sizeof(acTime), "%Y-%m-%d %H:%M:%S", localtime(&lStart));
        Output(0, "  start time: %s\n", acTime);
    }
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';   /* skip user time */
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';   /* skip system time */
    pcValue = pc;
    while (*pc && *pc != ' ') ++pc; *pc++ = '\0';
    if (*pcValue) Output(0, "  wait channel: %s\n", pcValue);
    pcValue = pc;
    while (*pc && *pc != '\n') ++pc;
    while (*pc && *pc == '\n') *pc-- = '\0';
    if (*pcValue) Output(0, "  credentials: %s\n", pcValue);
}
#endif

#ifdef OS_SOLARIS
void PrintStatusInfoSolaris(pid_t tuPID)
{
    struct prpsinfo soInfo;
    char acBuf[2048];
    char acTime[128];
    char* aszEnv[512];
    char* pszDesc;
    int i = 0;
    int nFD;

    snprintf(acBuf, sizeof(acBuf), "/proc/%u", (unsigned)tuPID);
    if ((nFD = open(acBuf, O_RDONLY)) == -1) return;
    if (ioctl(nFD, PIOCPSINFO, &soInfo) != -1)
    {
        Output(0, "  command: %s\n", soInfo.pr_fname);
        Output(0, "  cmdline: %s\n", soInfo.pr_psargs);
        pszDesc = GetStateDesc(soInfo.pr_sname);
        if (pszDesc) Output(0, "  state: %c - %s\n", soInfo.pr_sname, pszDesc);
        else Output(0, "  unknown state '%c'\n", soInfo.pr_sname);
        strftime(acTime, sizeof(acTime), "%Y-%m-%d %H:%M:%S", localtime(&soInfo.pr_start.tv_sec));
        Output(0, "  running since: %s\n", acTime);
        Output(0, "  uid: %u\n", soInfo.pr_uid);
        Output(0, "  gid: %u\n", soInfo.pr_gid);
        Output(0, "  ppid: %u\n", soInfo.pr_ppid);
        Output(0, "  sid: %u\n", soInfo.pr_sid);
        Output(0, "  process group: %u\n", soInfo.pr_pgrp);
        if (pread(nFD, aszEnv, sizeof(aszEnv), (off_t)soInfo.pr_envp) != -1)
        {
            while (i < 512 && aszEnv[i])
            {
                if (pread(nFD, acBuf, sizeof(acBuf), (off_t)aszEnv[i]) != -1)
                {
                    if (i == 0) Output(1, "  environment variables:\n");
                    Output(1, "  - %s\n", aszEnv[i]);
                    ++i;
                }
                else break;
            }
        }
    }
    close(nFD);
}
#endif

void PrintStatusInfo(pid_t tuPID)
{
    char acBuf[2048];
    FILE* fsStatus;

    snprintf(acBuf, sizeof(acBuf), "/proc/%u/status", (unsigned)tuPID);
    if ((fsStatus = fopen(acBuf, "r")))
    {
        while (!feof(fsStatus))
        {
            if (fgets(acBuf, sizeof(acBuf), fsStatus))
            {
                #ifdef OS_LINUX
                if (!strncmp(acBuf, "Uid:", 4) || (!strncmp(acBuf, "Gid:", 4))) Output(0, "  %s", acBuf);
                if (!strncmp(acBuf, "Groups:", 7))
                {
                    Output(0, "  %s", acBuf);
                    break;
                }
                #elif defined(OS_BSD)
                PrintStatusInfoBSD(acBuf);
                break;
                #endif
            }
        }
        fclose(fsStatus);
    }
}

void ListPrint(struct pid_list_t* tpsoList)
{
    struct pid_list_t* psoIter = tpsoList;
    char* pszDetectionType;
    #ifdef OS_LINUX
    char acPath[128];
    struct stat soStatBuf;
    FILE* fsFile;
    char acBuf[1024];
    ssize_t nLen;

    GetBootTime();
    #endif
    while (psoIter)
    {
        switch (psoIter->nType)
        {
            case SELFTEST:
                pszDetectionType = "selftest via --fake-hidden";
                break;
            case HIDDEN:
                pszDetectionType = "detected via system calls";
                break;
            case BRUTE:
                pszDetectionType = "detected via brute force";
                break;
            case FAKE:
                pszDetectionType = "faked process name";
                break;
            default:
                pszDetectionType = "unhandled detection type";
                assert(0);
        }
        Output(0, "\n* hidden pid %u (%s [%d])\n", (unsigned)psoIter->uPID, pszDetectionType, psoIter->nType);
        if (psoIter->nType != FAKE) CalcSignature(psoIter->uPID);
        #ifdef OS_LINUX
        PrintStatInfo(psoIter->uPID);
        #endif
        #if defined(OS_LINUX) || defined(OS_BSD)
        PrintStatusInfo(psoIter->uPID);
        #elif defined(OS_SOLARIS)
        PrintStatusInfoSolaris(psoIter->uPID);
        #endif
        #ifdef OS_LINUX
        /* /proc/pid/cmdline */
        GetCmdLine(psoIter->uPID, acBuf, sizeof(acBuf));
        if (acBuf[0] != '\0') Output(0, "  cmdline: '%s'\n", acBuf);
        /* /proc/pid/exe */
        snprintf(acPath, sizeof(acPath), "/proc/%u/exe", (unsigned)psoIter->uPID);
        if (stat(acPath, &soStatBuf) == 0)
        {
            if ((nLen = readlink(acPath, acBuf, sizeof(acBuf)-1)) > 0)
            {
                acBuf[nLen] = '\0';
                Output(0, "  exe: '%s'\n", acBuf);
            }
        }
        /* /proc/pid/wchan */
        snprintf(acPath, sizeof(acPath), "/proc/%u/wchan", (unsigned)psoIter->uPID);
        if (stat(acPath, &soStatBuf) == 0)
        {
            if ((fsFile = fopen(acPath, "r")))
            {
                if (fgets(acBuf, sizeof(acBuf), fsFile))
                    Output(0, "  wchan: '%s'\n", acBuf);
                fclose(fsFile);
            }
        }
        /* /proc/pid/environ */
        PrintEnviron(psoIter->uPID);
        #endif
        psoIter = psoIter->psoNext;
    }
}

void Scan(const char* tcszAction)
{
    if (opt_nVerbose)
    {
        char acBuf[128];
        time_t lScanTime = time(NULL);
        struct tm* psoTime = localtime(&lScanTime);

        strftime(acBuf, sizeof(acBuf), "%Y-%m-%d %H:%M:%S", psoTime);
        Output(1, "Scan %s at:%s %s\n", tcszAction, opt_nVerbose == 1 && strlen(tcszAction) == 5 ? "  " : "", acBuf);
    }
}

/************* Tests *************
 * return 1 for visible process  *
 * return 0 for no process found *
 *********************************/

/* system function tests (fast) */

void Test_getpriority(pid_t tuPID)
{
    errno = 0;
    getpriority(PRIO_PROCESS, tuPID);
    if (errno == 0) ++g_uVisible;
    #ifndef OS_SOLARIS
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
    #endif
}

void Test_getpgid(pid_t tuPID)
{
    if (getpgid(tuPID) != -1) ++g_uVisible;
    else ++g_uHidden;
}

void Test_getsid(pid_t tuPID)
{
    if (getsid(tuPID) != -1) ++g_uVisible;
    else ++g_uHidden;
}

void Test_kill(pid_t tuPID)
{
    if (kill(tuPID, 0) != -1) ++g_uVisible;
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
}

#ifdef HAS_SIGQUEUE
void Test_sigqueue(pid_t tuPID)
{
    union sigval soValue;
    soValue.sival_int = 0;
    if (sigqueue(tuPID, 0, soValue) != -1) ++g_uVisible;
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
}
#endif

#ifdef HAS_TKILL
void Test_tkill(pid_t tuPID)
{
    if (syscall(SYS_tkill, tuPID, 0) != -1) ++g_uVisible;
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
}
#endif

#ifdef HAS_SCHEDULER_H
#ifdef SCHED_GETAFFINITY
void Test_sched_getaffinity(pid_t tuPID)
{
    cpu_set_t soMask;
    errno = 0;
    #ifdef SCHED_GETAFFINITY2
    sched_getaffinity(tuPID, &soMask);
    #else
    sched_getaffinity(tuPID, sizeof(soMask), &soMask);
    #endif
    if (errno == 0) ++g_uVisible;
    else ++g_uHidden;
}
#endif

void Test_sched_getparam(pid_t tuPID)
{
    struct sched_param soParam;
    errno = 0;
    sched_getparam(tuPID, &soParam);
    if (errno == 0) ++g_uVisible;
    #ifndef OS_SOLARIS
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
    #endif
}

void Test_sched_rr_get_interval(pid_t tuPID)
{
    struct timespec soTime;
    errno = 0;
    sched_rr_get_interval(tuPID, &soTime);
    if (errno == 0) ++g_uVisible;
    #ifndef OS_SOLARIS
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
    #endif
}

void Test_sched_getscheduler(pid_t tuPID)
{
    if (sched_getscheduler(tuPID) != -1) ++g_uVisible;
    #ifndef OS_SOLARIS
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno != ESRCH) ++g_uVisible;
    else ++g_uHidden;
    #endif
}
#endif

#ifdef HAS_CAPABILITY
void Test_capgetp(pid_t tuPID)
{
    cap_t soCap;
    errno = 0;
    soCap = cap_init();
    if (soCap) capgetp(tuPID, soCap);
    if (!soCap)
    {
        Output(0, "Can't allocate enough memory: %s\n", strerror(errno));
        exit(103);
    }
    if (errno == ESRCH) ++g_uHidden;
    else ++g_uVisible;
    if (soCap) cap_free(soCap);
}
#endif

/* /proc tests (slower) */

void Test_chdir(pid_t tuPID)
{
    char acPath[128];
    #if defined(OS_LINUX) && !defined(USE_OLD_PS)
    pid_t uTGID;
    #endif

    snprintf(acPath, sizeof(acPath), "/proc/%u", (unsigned)tuPID);
    if (chdir(acPath) == 0)
    {
        ++g_uVisible;
        #if defined(OS_LINUX) && !defined(USE_OLD_PS)
        if ((uTGID = GetTGID(tuPID)))
        {
            ++g_uVisible;
            if (g_soKernel.bHasProcTask && opt_tgid)
            {
                snprintf(acPath, sizeof(acPath), "/proc/%u/task/%u", uTGID, tuPID);
                if (chdir(acPath) == 0) ++g_uVisible;
                else if (CheckZombie(tuPID)) ++g_uHidden;
            }
        }
        else ++g_uHidden;
        #endif
    }
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno == ENOENT) ++g_uHidden;
    else ++g_uVisible;
}

void Test_opendir(pid_t tuPID)
{
    DIR* psoDir;
    char acPath[128];

    snprintf(acPath, sizeof(acPath), "/proc/%u", (unsigned)tuPID);
    psoDir = opendir(acPath);
    if (psoDir)
    {
        ++g_uVisible;
        closedir(psoDir);
        #if defined(OS_LINUX) && !defined(USE_OLD_PS)
        /* GetTGID already checked if opt_chdir is set */
        if (!(opt_chdir && !opt_tgid))
        {
            pid_t tgid = GetTGID(tuPID);

            if (tgid)
            {
                ++g_uVisible;
                if (g_soKernel.bHasProcTask && opt_tgid)
                {
                    snprintf(acPath, sizeof(acPath), "/proc/%u/task/%u", tgid, tuPID);
                    errno = 0;
                    psoDir = opendir(acPath);
                    if (psoDir)
                    {
                        ++g_uVisible;
                        closedir(psoDir);
                    }
                    else if (CheckZombie(tuPID)) ++g_uHidden;
                }
            }
            else ++g_uHidden;
        }
        #endif
    }
    else if (g_uUID == 0) ++g_uHidden;
    else if (errno == ENOENT) ++g_uHidden;
    else ++g_uVisible;
}

void GetProcAll(void)
{
    struct dirent* psoProcEnt;
    DIR*     psoProcDir;
    unsigned uPID;
    #ifdef OS_LINUX
    struct dirent* psoTaskEnt;
    DIR*     psoTaskDir;
    unsigned uTaskPID;
    char     acTask[128];
    char     acCheckPID[128];
    #endif

    psoProcDir = opendir("/proc");
    if (!psoProcDir)
    {
        Output(0, "Can't open /proc directory: %s", strerror(errno));
        exit(102);
    }
    
    if (g_payProcPIDs) free(g_payProcPIDs);
    g_payProcPIDs = BitmapInit();
    while ((psoProcEnt = readdir(psoProcDir)))
    {
        if (isdigit((int)*psoProcEnt->d_name))
        {
            sscanf(psoProcEnt->d_name, "%u", &uPID);
            BitmapSet(g_payProcPIDs, uPID);
            #ifdef OS_LINUX
            if (g_soKernel.bHasProcTask)
            {
                snprintf(acTask, sizeof(acTask), "/proc/%s/task", psoProcEnt->d_name);
                psoTaskDir = opendir(acTask);
                if (!psoTaskDir)
                {
                    /* skip special files (no pid directory) */
                    snprintf(acCheckPID, sizeof(acCheckPID), "%u", uPID);
                    if (strncmp(acCheckPID, psoProcEnt->d_name, sizeof(acCheckPID))) continue;
                    /* process without task dir */
                    Output(3, "Can't open %s (before double check).\n", acTask);
                }
                else
                {
                    while ((psoTaskEnt = readdir(psoTaskDir)))
                    {
                        if (!strncmp(psoTaskEnt->d_name, ".", 2) || !strncmp(psoTaskEnt->d_name, "..", 3)) continue;
                        sscanf(psoTaskEnt->d_name, "%u", &uTaskPID);
                        BitmapSet(g_payProcPIDs, uTaskPID);
                    }
                    closedir(psoTaskDir);
                }
            }
            #endif
        }
    }
    closedir(psoProcDir);
}

/* pretty slow - only for double checks (GetProcAll for all pids) */
void Test_readdir(pid_t tuPID)
{
    DIR* psoProc;
    struct dirent* psoProcEnt;
    char acPID[16];
    int bFound = 0;
    #ifdef OS_LINUX
    char acTaskPath[128];
    DIR* psoTaskDir;
    char acTaskPID[128];
    #endif

    psoProc = opendir("/proc");
    if (!psoProc)
    {
        Output(0, "Can't open /proc directory: %s", strerror(errno));
        exit(102);
    }
    
    snprintf(acPID, sizeof(acPID), "%u", (unsigned)tuPID);
    while ((psoProcEnt = readdir(psoProc)))
        if (strncmp(acPID, psoProcEnt->d_name, sizeof(acPID)) == 0)
        {
            ++bFound;
            ++g_uVisible;
            break;
        }
    closedir(psoProc);

    /* maybe this pid is a thread... */
    #ifdef OS_LINUX
    if (!bFound)
    {
        if (g_soKernel.bHasProcTask)
        {
            snprintf(acTaskPath, sizeof(acTaskPath), "/proc/%d/task", tuPID);
            psoProc = opendir(acTaskPath);
            if (!psoProc) ++g_uHidden;
            else
            {
                while (!bFound && (psoProcEnt = readdir(psoProc)))
                {
                    if (!strncmp(psoProcEnt->d_name, ".", 2) || !strncmp(psoProcEnt->d_name, "..", 3) || !strncmp(psoProcEnt->d_name, acPID, sizeof(acPID))) continue;
                    snprintf(acTaskPID, sizeof(acTaskPID), "/proc/%s/task/%d", psoProcEnt->d_name, tuPID);
                    psoTaskDir = opendir(acTaskPID);
                    if (psoTaskDir)
                    {
                        ++g_uVisible;
                        ++bFound;
                        closedir(psoTaskDir);
                    }
                }
                closedir(psoProc);
                if (!bFound) ++g_uHidden;
            }
        }
        #if !defined(USE_OLD_PS)
        else if (tuPID == GetTGID(tuPID)) ++g_uHidden;
        #endif
    }
    #endif
}

void Test_proc(pid_t tuPID)
{
    char acPath[32];
    struct stat soStat;

    snprintf(acPath, sizeof(acPath), "/proc/%u", (unsigned)tuPID);
    if (stat(acPath, &soStat) != -1) ++g_uVisible;
    else ++g_uHidden;
}

/* /bin/ps tests */

void GetAllOfPS(void)
{
    char     acLine[256];
    FILE*    fsPS;
    unsigned uPID;

    /* avoid getcwd errors while inside of /proc/XXX of terminated process */
    if (chdir("/") != 0) Output(3, "Can't chdir /: %s", strerror(errno));
    if (!(fsPS = popen(PS_ALL, "r")))
    {
        Output(0, "Calling ps failed: %s\n", strerror(errno));
        exit(105);
    }
    if (g_payPS_PIDs) free(g_payPS_PIDs);
    g_payPS_PIDs = BitmapInit();
    while (fgets(acLine, sizeof(acLine), fsPS))
    {
        sscanf(acLine, "%u", &uPID);
        BitmapSet(g_payPS_PIDs, uPID);
    }
    if (pclose(fsPS) == -1)
        Output(3, "Can't close ps command '%s': %s\n", PS_ALL, strerror(errno));
}

struct pid_list_t* GetListBruteHidden(int tnType)
{
    struct pid_list_t* psoList = NULL;
    struct pid_list_t** ppsoIter = &psoList;
    pid_t uPID;
    unsigned uHidden = 0;

    /* PIDs below 300 are reserved for kernel - see RESERVED_PIDS in linux/kernel/pid.c */
    for (uPID = g_uReservedPIDs; uPID < g_uMaxPID; ++uPID)
    {
        if (!BitmapGet(g_payPS_PIDs, uPID))
        {
            ListAppend(ppsoIter, uPID, tnType);
            ++uHidden;
            ppsoIter = &((*ppsoIter)->psoNext);
        }
    }
    /* fast forking processes, many "hidden" pids, activate random pid checking */
    if (uHidden > 1000) g_soKernel.bHasRandomPIDs = 1;
    return psoList;
}

/* brute force tests */
struct pid_list_t* TestBrute_vfork(void)
{
    pid_t uStart = 0;
    pid_t uPID = 1;
    unsigned long i = 1;
    unsigned long ulMaxIterations = g_uMaxPID * (1 + g_soKernel.bHasRandomPIDs + opt_nRandom + g_nMoreBruteTesting);
    int nStatus;
    #ifdef DEBUG
    pid_t uLowPID = g_uMaxPID;
    pid_t uHighPID = 0;
    #endif

    SetupCycleWatchdog();
    Output(2, "");
    GetAllOfPS();
    for (errno = 0; (uStart != uPID || g_soKernel.bHasRandomPIDs || g_nMoreBruteTesting) && i < ulMaxIterations && !errno; ++i)
    {
        if ((uPID = vfork()) == 0) _exit(0);
        else
        {
            BitmapSet(g_payPS_PIDs, uPID);
            #ifdef DEBUG
            if (uPID > uHighPID) uHighPID = uPID;
            else if (uPID < uLowPID) uLowPID = uPID;
            #endif
            waitpid(uPID, &nStatus, 0);
            if (!uStart)
            {
                uStart = uPID;
                uPID = 0;
            }
        }
    }

    if (errno)
    {
        Output(0, "Can't vfork: %s\n", strerror(errno));
        exit(106);
    }
    Output(3, "Debug: forked %u times out of %u pid_max\n", i, g_uMaxPID);
    #ifdef DEBUG
    Output(3, "Debug: min pid %u / max pid %u\n", uLowPID, uHighPID);
    #endif
    return GetListBruteHidden(BRUTE);
}

#if defined(HAS_PTHREAD) && defined(OS_LINUX)
void* ThreadBody(void* tpvPID)
{
    #ifdef SYS_gettid
    *(pid_t*)tpvPID = (pid_t)syscall(SYS_gettid);
    #else
    *(pid_t*)tpvPID = (pid_t)syscall(SYS_getpid);
    #endif
    return tpvPID;
}

struct pid_list_t* TestBrute_pthread(void)
{
    pid_t uStart = 0;
    pid_t uPID = 1;
    int nErr;
    pthread_t soThread;
    unsigned long i = 1;
    unsigned long ulMaxIterations = g_uMaxPID * (1 + g_soKernel.bHasRandomPIDs + opt_nRandom + g_nMoreBruteTesting);
    #ifdef DEBUG
    pid_t nLowPID = g_uMaxPID;
    pid_t nHighPID = 0;
    #endif

    SetupCycleWatchdog();
    GetAllOfPS();
    for (; (uStart != uPID || g_soKernel.bHasRandomPIDs || g_nMoreBruteTesting) && i < ulMaxIterations; ++i)
    {
        if ((nErr = pthread_create(&soThread, NULL, ThreadBody, &uPID)))
        {
            Output(0, "Can't create thread: %s\n", strerror(nErr));
            exit(106);
        }
        if ((nErr = pthread_join(soThread, NULL)))
        {
            Output(0, "Can't join thread: %s\n", strerror(nErr));
            exit(106);
        }
        BitmapSet(g_payPS_PIDs, uPID);
        #ifdef DEBUG
        if (uPID > nHighPID) nHighPID = uPID;
        else if (uPID < nLowPID) nLowPID = uPID;
        #endif
        if (!uStart)
        {
            uStart = uPID;
            uPID = 0;
        }
    }
    Output(3, "Debug: created %u threads out of %u pid_max\n", i, g_uMaxPID);
    #ifdef DEBUG
    Output(3, "Debug: min pid %u / max pid %u\n", nLowPID, nHighPID);
    #endif
    return GetListBruteHidden(BRUTE);
}
#endif

void CalcSignature(pid_t tuPID)
{
    #define MK_SIG(test) { test(tuPID); if (g_uHidden != uOldHidden) { ulSignature |= i; uOldHidden = g_uHidden; } } ulAllFailed |= i; i <<= 1
    char* pszRootKit = NULL;
    unsigned long i = 1;
    unsigned long ulSignature = 0;
    unsigned long ulAllFailed = 0;
    unsigned uOldHidden = 0;

    if (opt_nOmitSignature) return;
    if (g_uUID != 0) Output(2, "  rootkit signature detection should run with root permission\n");
    g_uVisible = g_uHidden = 0;
    if (!BitmapGet(g_payPS_PIDs, tuPID)) ulSignature |= i; ulAllFailed |= i; i <<= 1;
    #ifdef DEBUG
    if (opt_getpgid)
    #endif
    MK_SIG(Test_getpgid);
    #ifdef DEBUG
    if (opt_getpriority)
    #endif
    MK_SIG(Test_getpriority);
    #ifdef DEBUG
    if (opt_getsid)
    #endif
    MK_SIG(Test_getsid);
    #ifdef DEBUG
    if (opt_kill)
    #endif
    MK_SIG(Test_kill);
    #ifdef HAS_SIGQUEUE
    #ifdef DEBUG
    if (opt_sigqueue)
    #endif
    MK_SIG(Test_sigqueue);
    #else
    ulAllFailed |= 1; i <<= 1;
    #endif
    #ifdef DEBUG
    if (opt_readdir)
    #endif
    MK_SIG(Test_readdir);
    #ifdef DEBUG
    if (opt_chdir)
    #endif
    MK_SIG(Test_chdir);   
    #ifdef DEBUG
    if (opt_opendir)
    #endif
    MK_SIG(Test_opendir);
    #ifdef DEBUG
    if (opt_proc)
    #endif
    MK_SIG(Test_proc);
    #ifdef HAS_TKILL
    #ifdef DEBUG
    if (opt_tkill)
    #endif
    MK_SIG(Test_tkill);
    #else
    ulAllFailed |= 1; i <<= 1;
    #endif
    #ifdef HAS_SCHEDULER_H
    #ifdef DEBUG
    if (opt_sched_getparam)
    #endif
    MK_SIG(Test_sched_getparam);
    #ifdef DEBUG
    if (opt_sched_rr_get_interval)
    #endif
    MK_SIG(Test_sched_rr_get_interval);
    #ifdef DEBUG
    if (opt_sched_getscheduler)
    #endif
    MK_SIG(Test_sched_getscheduler);
    #else
    ulAllFailed |= 7; i <<= 3;
    #endif
    #ifdef SCHED_GETAFFINITY
    #ifdef DEBUG
    if (opt_sched_getaffinity)
    #endif
    MK_SIG(Test_sched_getaffinity);
    #else
    ulAllFailed |= 1; i <<= 1;
    #endif
    #ifdef HAS_CAPABILITY
    #ifdef DEBUG
    if (opt_capgetp)
    #endif
    MK_SIG(Test_capgetp);
    #else
    ulAllFailed |= 1; i <<= 1;
    #endif
    if (!ulSignature && opt_nFakeHidden && tuPID == getpid()) return;
    else if ((ulSignature|1) == ulAllFailed) pszRootKit = "false positive (fast forking processes or random pids?)";
    #ifdef OS_LINUX
    else switch (ulSignature)
    {
        case     1: pszRootKit = "kBeast-v1"; break; /* hidden against ps */
        case    16: pszRootKit = "kBeast-v1"; break; /* hidden against: kill, but visible with "ps ... | cat" */
        case    17: if (!strcmp(g_soKernel.pszName, "Linux") && g_soKernel.nMajor <= 4 && g_soKernel.nRelease == 2) pszRootKit = "SucKIT";
                    break;
        case    65: pszRootKit = "Snakso.a"; break; /* hidden against: readdir */
        case   833: pszRootKit = "Jynx-Kit"; break; /* hidden against: readdir opendir proc */
        case   981: pszRootKit = "Phalanx2"; break; /* hidden against: getpriority kill readdir chdir opendir proc */
    }
    #endif
    if (pszRootKit) Output(0, "  possibly %s [sig: %lu]\n", pszRootKit, ulSignature);
    else if (g_uUID == 0) Output(0, "  unknown rootkit signature [%lu] - please contact %s\n", ulSignature, MAILADDR);
    else Output(0, "  unknown rootkit signature [%lu] - please run %s as root\n", ulSignature, g_szProgramName);
}

void DoBruteTest(struct pid_list_t** tppsoListHidden, struct pid_list_t* (*fnBruteTest)(void), const char* tszName)
{
    struct pid_list_t* psoListBrute = NULL;
    struct pid_list_t* psoListTmp = NULL;

    BruteTest(&psoListBrute, fnBruteTest, tszName);
    if (psoListBrute)
    {
        psoListTmp = ListUnion(*tppsoListHidden, psoListBrute);
        ListClear(*tppsoListHidden);
        ListClear(psoListBrute);
        *tppsoListHidden = psoListTmp;
    }
}

void DoTests(pid_t tuPID, int bDoubleCheck)
{
    g_uVisible = g_uHidden = 0;
    /* evaluate ps output */
    if (BitmapGet(g_payPS_PIDs, tuPID)) ++g_uVisible;
    else ++g_uHidden;
    /* system tests */
    if (opt_getpgid)                Test_getpgid(tuPID);
    if (opt_getpriority)            Test_getpriority(tuPID);
    if (opt_getsid)                 Test_getsid(tuPID);
    if (opt_kill)                   Test_kill(tuPID);
    #ifdef HAS_SIGQUEUE
    if (opt_sigqueue)               Test_sigqueue(tuPID);
    #endif
    #ifdef HAS_TKILL
    if (opt_tkill)                  Test_tkill(tuPID);
    #endif
    #ifdef HAS_SCHEDULER_H
    if (opt_sched_getparam)         Test_sched_getparam(tuPID);
    if (opt_sched_rr_get_interval)  Test_sched_rr_get_interval(tuPID);
    if (opt_sched_getscheduler)     Test_sched_getscheduler(tuPID);
    #ifdef SCHED_GETAFFINITY
    if (opt_sched_getaffinity)      Test_sched_getaffinity(tuPID);
    #endif
    #endif
    #ifdef HAS_CAPABILITY
    if (opt_capgetp)                Test_capgetp(tuPID);
    #endif
    /* proc tests */
    if (opt_readdir)
    {
        if (bDoubleCheck)
        {
            /* Test_readdir only for double check, far too slow */
            Test_readdir(tuPID);
        }
        else
        {
            if (BitmapGet(g_payProcPIDs, tuPID)) ++g_uVisible;
            else ++g_uHidden;
        }
    }
    if (opt_chdir)                  Test_chdir(tuPID);    /* MUST be called before Test_opendir */
    if (opt_opendir)                Test_opendir(tuPID);  /* MUST be called after Test_chdir */
    if (opt_proc)                   Test_proc(tuPID);
}

void MakeFinalTest(struct pid_list_t** tppsoListResult)
{
    struct pid_list_t* psoListTmp = NULL;
    struct pid_list_t* psoListBrute = NULL;
    struct pid_list_t* psoListHidden = NULL;
    struct pid_list_t* psoListDoubleCheck = NULL;

    Output(2, "Doing more brute force testing to avoid false positives...\n");
    g_nMoreBruteTesting = 5;
    SearchBruteForce(&psoListDoubleCheck);
    ListSplitBrute(tppsoListResult, &psoListBrute, &psoListHidden);
    psoListTmp = ListIntersect(psoListBrute, psoListDoubleCheck);
    ListClear(psoListBrute);
    ListClear(psoListDoubleCheck);
    *tppsoListResult = ListUnion(psoListTmp, psoListHidden);
    ListClear(psoListTmp);
    ListClear(psoListHidden);
    if (opt_nDaemon) g_nMoreBruteTesting = 0;
}

int NeedMoreBruteForceChecking(struct pid_list_t* tpsoListHidden)
{
    struct pid_list_t* psoIter = tpsoListHidden;

    while (psoIter)
    {
        if (psoIter->nType == BRUTE) return 1;
        psoIter = psoIter->psoNext;
    }
    return 0;
}

void PrintTests(void)
{
    printf("+ ps output             (always)\n");
    printf("%c readdir\n", opt_readdir ? '+' : '-');
    printf("%c chdir\n", opt_chdir ? '+' : '-');
    printf("%c opendir\n", opt_opendir ? '+' : '-');
    printf("%c proc\n", opt_proc ? '+' : '-');
    printf("%c getpgid\n", opt_getpgid ? '+' : '-');
    printf("%c getpriority\n", opt_getpriority ? '+' : '-');
    printf("%c getsid\n", opt_getsid ? '+' : '-');
    printf("%c kill\n", opt_kill ? '+' : '-');
    #ifdef HAS_SIGQUEUE
    printf("%c sigqueue\n", opt_sigqueue ? '+' : '-');
    #elif !defined(OS_BSD)
    printf("_ sigqueue              (not build)\n");
    #elif DEBUG
    printf("_ sigqueue              (unavailable)\n");
    #endif
    #ifdef HAS_TKILL
    printf("%c tkill\n", opt_tkill ? '+' : '-');
    #elif defined(OS_LINUX) && !defined(USE_OLD_PS)
    printf("_ tkill                 (not build)\n");
    #elif DEBUG
    printf("_ tkill                 (unavailable)\n");
    #endif
    #ifdef HAS_SCHEDULER_H
    printf("%c sched_getparam\n", opt_sched_getparam ? '+' : '-');
    printf("%c sched_rr_get_interval\n", opt_sched_rr_get_interval ? '+' : '-');
    printf("%c sched_getscheduler\n", opt_sched_getscheduler ? '+' : '-');
    #ifdef SCHED_GETAFFINITY
    printf("%c sched_getaffinity\n", opt_sched_getaffinity ? '+' : '-');
    #elif DEBUG
    printf("_ sched_getaffinity     (unavailable)\n");
    #endif
    #else
    printf("_ sched_getparam        (not build)\n");
    printf("_ sched_rr_get_interval (not build)\n");
    printf("_ sched_getscheduler    (not build)\n");
    printf("_ sched_getaffinity     (not build)\n");
    #endif
    #ifdef HAS_CAPABILITY
    printf("%c capgetp\n", opt_capgetp ? '+' : '-');
    #elif defined(OS_LINUX)
    printf("_ capgetp               (not build)\n");
    #elif DEBUG
    printf("_ capgetp               (unavailable)\n");
    #endif
    printf("%c vfork                 (brute force)\n", opt_vfork ? '+' : '-');
    #ifdef HAS_PTHREAD
    #ifdef OS_LINUX
    printf("%c pthread               (brute force)\n", opt_pthread ? '+' : '-');
    #else
    printf("_ pthread               (not build)\n");
    #endif
    #elif DEBUG
    printf("_ pthread               (pointless)\n");
    #endif
    #if defined(OS_LINUX) || defined(OS_SOLARIS)
    printf("%c fake_name\n", opt_fake_name ? '+' : '-');
    #elif DEBUG
    printf("_ fake_name             (pointless)\n");
    #endif
}

int SearchBruteForce(struct pid_list_t** tppsoListHidden)
{
    if (opt_vfork)      DoBruteTest(tppsoListHidden, TestBrute_vfork, "vfork");
    #if defined(HAS_PTHREAD) && defined(OS_LINUX)
    if (opt_pthread)    DoBruteTest(tppsoListHidden, TestBrute_pthread, "pthread");
    #endif
    return ListLen(*tppsoListHidden);
}

#if defined(OS_LINUX) || defined(OS_SOLARIS)
int SearchFakeNames(struct pid_list_t** tppsoListHidden)
{
    #ifdef OS_SOLARIS
    struct prpsinfo soInfo;
    char* aszEnv[512];
    char acEnvBuf[1024];
    int bFound;
    int nFD;
    int nSpaces = 60 - opt_nFakeSearch * 15;
    #else
    int nSpaces = 100 - opt_nFakeSearch * 25;
    #endif
    #if defined(OS_LINUX) || defined(OS_SOLARIS)
    int i = 0;
    #endif
    unsigned uHidden = 0;
    char acPath[PATH_MAX];
    char* pcCmdLine;
    DIR* psoProcDir;
    struct dirent* psoProcEnt;
    unsigned uPID;
    struct stat soStat;

    Output(2, "Searching for faked process names.\n");
    SetupCycleWatchdog();
    if (nSpaces <= 0) nSpaces = 10;
    if (!(psoProcDir = opendir("/proc")))
    {
        Output(0, "Can't open /proc directory: %s", strerror(errno));
        exit(102);
    }
    while ((psoProcEnt = readdir(psoProcDir)))
    {
        if (isdigit((int)*psoProcEnt->d_name))
        {
            sscanf(psoProcEnt->d_name, "%u", &uPID);
            if ((pcCmdLine = CheckProcessName(uPID, nSpaces)))
            {
                if (opt_nFakeSearch <= 2)
                {
                    #if defined(OS_LINUX)
                    snprintf(acPath, sizeof(acPath), "/proc/%u/exe", uPID);
                    /* search only for executables in hidden directories unless opt_nFakeSearch > 1 */
                    if ((opt_nFakeSearch <= 1) && ((i = readlink(acPath, acPath, sizeof(acPath)-1)) > 0))
                    {
                        acPath[i] = '\0';
                        if (!strstr(acPath, "/.")) continue;
                    }
                    #elif defined(OS_SOLARIS)
                    if (opt_nFakeSearch <= 1)
                    {
                        snprintf(acPath, sizeof(acPath), "/proc/%u", uPID);
                        if ((nFD = open(acPath, O_RDONLY)) == -1) continue;
                        if (ioctl(nFD, PIOCPSINFO, &soInfo) != -1)
                        {
                            if (!strstr(soInfo.pr_psargs, "/."))
                            {
                                bFound = i = 0;
                                if (pread(nFD, aszEnv, sizeof(aszEnv), (off_t)soInfo.pr_envp) != -1)
                                {
                                    while (!bFound && i < 512 && aszEnv[i])
                                    {
                                        if (pread(nFD, acEnvBuf, sizeof(acEnvBuf), (off_t)aszEnv[i]) != -1)
                                        {
                                            if (strncmp(acEnvBuf, "PWD=", 4) == 0)
                                            {
                                                if (strstr(acEnvBuf, "/.")) ++bFound;
                                                else break;
                                            }
                                            ++i;
                                        }
                                        else break;
                                    }
                                }
                                if (!bFound)
                                {
                                    close(nFD);
                                    continue;
                                }
                            }
                        }
                        close(nFD);
                    }
                    #endif
                    /* ignore root processes unless opt_nFakeSearch > 2 */
                    snprintf(acPath, sizeof(acPath), "/proc/%u", uPID);
                    if ((opt_nFakeSearch <= 2) && (stat(acPath, &soStat) == 0) && !soStat.st_uid) continue;
                }
                ++uHidden;
                ListAppend(tppsoListHidden, uPID, FAKE);
            }
        }
    }
    closedir(psoProcDir);
    Output(2, "Found %d faked process name%s.\n", uHidden, (uHidden > 1) ? "s" : "");
    return uHidden;
}
#endif

int SearchHidden(void)
{
    struct pid_list_t* psoListHidden = NULL;
    struct pid_list_t* psoListDoubleCheck = NULL;
    struct pid_list_t* psoListIter = NULL;
    struct pid_list_t* psoListIterDC = NULL;
    char acCurDir[PATH_MAX];
    pid_t uMyPID = getpid();
    pid_t uPID = 1;
    unsigned uHidden = 0;
    int i;

    if (g_uGlobalTimeoutSecs && !g_uCycleTimeoutSecs)
    {
        signal(SIGALRM, GlobalWatchdog);
        alarm(g_uGlobalTimeoutSecs);
    }

    SetupCycleWatchdog();
    /* for chdir test, keep current working directory */
    if ((opt_chdir || opt_opendir) && !getcwd(acCurDir, sizeof(acCurDir)))
    {
        snprintf(acCurDir, sizeof(acCurDir), "/");
        Output(1, "Can't get current working directory: %s\n", strerror(errno));
    }

    GetMaxPID();
    Scan("started");
    GetAllOfPS();
    if (opt_readdir) GetProcAll();
    for (; uPID < g_uMaxPID; ++uPID)
    {
        if (uPID == uMyPID) continue;
        DoTests(uPID, 0);
        if (g_uVisible && g_uHidden)
        {
            ListAppend(&psoListIter, uPID, HIDDEN);
            if (!psoListHidden) psoListHidden = psoListIter;
        }
    }
    uHidden = ListLen(psoListHidden);
    Output(2, "Found %u hidden process%s after step 1 (at least one or two of ps command).\n", uHidden, (uHidden > 1) ? "es" : "");

    for (i = 2; uHidden && i <= g_nMaxChecks; ++i)
    {
        SetupCycleWatchdog();
        /* double check of suspicious pids */
        psoListIter = psoListHidden;
        GetAllOfPS();
        while (psoListIter)
        {
            Output(3, "Debug: double check pid %u\n", psoListIter->uPID);
            DoTests(psoListIter->uPID, 1);
            if (g_uVisible && g_uHidden)
            {
                ListAppend(&psoListIterDC, psoListIter->uPID, psoListIter->nType);
                if (!psoListDoubleCheck) psoListDoubleCheck = psoListIterDC;
            }
            psoListIter = psoListIter->psoNext;
        }
        uHidden = ListLen(psoListDoubleCheck);
        Output(2, "Found %u hidden process%s after step %i.\n", uHidden, (uHidden > 1) ? "es" : "", i);
        if (ListCmp(psoListHidden, psoListDoubleCheck))
        {
            /* lists are not equal */
            ListClear(psoListHidden);
            psoListHidden = psoListDoubleCheck;
            psoListDoubleCheck = psoListIterDC = NULL;
        }
        else
        {
            ListClear(psoListDoubleCheck);
            break;
        }
    }

    /* skip brute force tests if there are already hidden pids */
    if (!psoListHidden || opt_nForceBrute)
        uHidden = SearchBruteForce(&psoListHidden);

    #if defined(OS_LINUX) || defined(OS_SOLARIS)
    /* search for processes with fake names */
    if (opt_fake_name)
        uHidden += SearchFakeNames(&psoListHidden);
    #endif

    /* back to cwd, if it was changed by chdir */
    if ((opt_chdir || opt_opendir) && chdir(acCurDir)) ;

    if (opt_nFakeHidden)
    {
        ListAppend(&psoListHidden, getpid(), SELFTEST);
        ++uHidden;
    }

    if (NeedMoreBruteForceChecking(psoListHidden))
    {
        MakeFinalTest(&psoListHidden);
        uHidden = ListLen(psoListHidden);
    }

    Scan("ended");
    Output(1, "\n");

    if (uHidden)
    {
        Output(0, "Found %u hidden process%s.\n", uHidden, (uHidden > 1) ? "es" : "");
        ListPrint(psoListHidden);
    }
    else if (opt_nVerbose) Output(1, "OK - no hidden processes found.\n");
    ListClear(psoListHidden);
    return (uHidden > 99) ? 100 : uHidden;
}

void InvalidOption(const char* tcszOption, int tnErr)
{
    switch(tnErr)
    {
        case  1: fprintf(stderr, "%s: option requires an argument -- '%s'\n", g_szProgramName, tcszOption); break;
        default: fprintf(stderr, "%s: invalid option -- '%s'\n", g_szProgramName, tcszOption); break;
    }
    fprintf(stderr, "Try `%s --help' for more information.\n", g_szProgramName);
    exit(101);
}

void SetOption(const char* tcszName, int* tpnOption)
{
    if (*tcszName == '+')
    {
        if (*tpnOption) Output(2, "Info: Option '%s' already set.\n", tcszName+1);
        *tpnOption = 1;
    }
    else if (*tcszName == '-')
    {
        if (!*tpnOption) Output(2, "Info: Option '%s' isn't set.\n", tcszName+1);
        *tpnOption = 0;
    }
    else InvalidOption(tcszName, INVALID);
    Output(3, "Debug: Option '%s' is%s set.\n", tcszName+1, *tpnOption ? "" : "n't");
}

/* setup watchdog between checks to kill long running or hanging tests */
void SetupCycleWatchdog(void)
{
    if (!g_uGlobalTimeoutSecs && g_uCycleTimeoutSecs)
    {
        signal(SIGALRM, CycleWatchdog);
        alarm(g_uCycleTimeoutSecs);
    }
}

void Usage(void)
{
    printf("Usage: %s [options] [+/-tests...]\n\n", g_szProgramName);
    printf("Options:\n");
    printf("   -b             brute force tests even if previous tests found hidden pid\n");
    printf("   -c <num>       retry checks up to <num> rounds (default: %d)\n", g_nMaxChecks);
    printf("   -f             search for process name fakes more intense\n");
    printf("   --fake-hidden  invoke a self-test\n");
    printf("   -h             display this help\n");
    printf("   -l             list all available tests and exit\n");
    #ifdef OS_LINUX
    printf("   -m <max>       use given max pids instead of %s\n", PATH_PID_MAX);
    #else
    printf("   -m <max>       use given max pids instead of auto-detected value\n");
    #endif
    printf("   -o             omit signature check\n");
    printf("   -p             set niceness to -20 (highest priority for process scheduling)\n");
    printf("   -q             do not print any messages to stdout\n");
    printf("   -r             system uses random pids, do more brute force forking\n");
    printf("   -s             log result also to syslog\n");
    printf("   -T <secs>      timeout for entire run after given seconds (default: %d)\n", g_uGlobalTimeoutSecs);
    printf("   -t <secs>      timeout between checks after given seconds (default: %d)\n", g_uCycleTimeoutSecs);
    printf("   -v             verbose\n");
    printf("   -w <file>      write output into given file\n");
    printf("   -V             show version and exit\n");
    #ifdef OS_LINUX
    printf("   -z             ignore suspicious zombie processes\n");
    #endif
}

int main(int argc, char *argv[])
{
    int i;
    char* pc; /* temp pointer */

    /* get basename of program name */
    g_szProgramName = pc = argv[0];
    while (*pc)
    {
        if (*pc == '/') g_szProgramName = pc+1;
        ++pc;
    }

    GetKernelInfo();
    for (i = 1; i < argc; ++i)
    {
        if ((strcmp(argv[i], "-h") == 0) || (strcmp(argv[i], "--help") == 0))
        {
            Usage();
            return 0;
        }
        else if ((strcmp(argv[i], "-V") == 0) || (strcmp(argv[i], "--version") == 0))
        {
            printf("%s version %s\n\n%s", g_szProgramName, VERSION, FOOTER);
            return 0;
        }
        else if ((strcmp(argv[i], "-b") == 0) || (strcmp(argv[i], "--force-brute") == 0))
        {
            ++opt_nForceBrute;
        }
        else if ((strcmp(argv[i], "-c") == 0) || (strcmp(argv[i], "--max-checks") == 0))
        {
            if (!argv[i+1]) InvalidOption(argv[i], MISSING);
            sscanf(argv[i+1], "%d", &g_nMaxChecks);
            if (!g_nMaxChecks) g_nMaxChecks = INT_MAX;
            Output(2, "Set max checks to %d.\n", g_nMaxChecks);
            ++i;
        }
        else if (strcmp(argv[i], "--fake-hidden") == 0)
        {
            ++opt_nFakeHidden;
        }
        else if ((strcmp(argv[i], "-l") == 0) || (strcmp(argv[i], "--list-tests") == 0))
        {
            ++opt_nListTests;
        }
        else if (strcmp(argv[i], "-m") == 0)
        {
            if (!(opt_szMaxPID = argv[++i])) InvalidOption(argv[i-1], MISSING);
        }
        else if (strcmp(argv[i], "-o") == 0)
        {
            ++opt_nOmitSignature;
        }
        else if (strcmp(argv[i], "-p") == 0)
        {
            ++opt_nNiceness;
        }
        else if ((strcmp(argv[i], "-q") == 0) || (strcmp(argv[i], "--quiet") == 0))
        {
            ++opt_nQuiet;
        }
        else if (strcmp(argv[i], "-s") == 0)
        {
            ++opt_nSyslog;
            openlog(g_szProgramName, LOG_PID, LOG_SYSLOG);
        }
        else if (strcmp(argv[i], "-T") == 0)
        {
            if (!argv[i+1]) InvalidOption(argv[i], MISSING);
            sscanf(argv[i+1], "%u", &g_uGlobalTimeoutSecs);
            if (g_uCycleTimeoutSecs && g_uCycleTimeoutSecs != DEFAULT_TIMEOUT)
            {
                fprintf(stderr, "You can't mix option -T with -t.\n");
                exit(101);
            }
            g_uCycleTimeoutSecs = 0;
            if (g_uGlobalTimeoutSecs) Output(2, "Timeout for entire run is set to %u seconds.\n", g_uGlobalTimeoutSecs);
            ++i;
        }
        else if (strcmp(argv[i], "-t") == 0)
        {
            if (!argv[i+1]) InvalidOption(argv[i], MISSING);
            sscanf(argv[i+1], "%u", &g_uCycleTimeoutSecs);
            if (g_uGlobalTimeoutSecs)
            {
                fprintf(stderr, "You can't mix option -t with -T.\n");
                exit(101);
            }
            if (!g_uCycleTimeoutSecs) Output(2, "Timeout between checks is disabled.\n");
            else if (!g_uCycleTimeoutSecs) Output(2, "Timeout is disabled.\n");
            else Output(2, "Timeout between checks is set to %u seconds.\n", g_uCycleTimeoutSecs);
            ++i;
        }
        else if ((strcmp(argv[i], "-w") == 0) || (strcmp(argv[i], "--write") == 0))
        {
            if (!argv[i+1]) InvalidOption(argv[i], MISSING);
            if (!(opt_fsLogFile = fopen(argv[i+1], "w")))
            {
                fprintf(stderr, "WARNING - Unable to open log file '%s': %s!\n", argv[i], strerror(errno));
            }
            ++i;
        }
        #ifdef OS_LINUX
        else if ((strcmp(argv[i], "-z") == 0) || (strcmp(argv[i], "--ignore-zombie") == 0))
        {
            ++opt_nIgnoreZombie;
        }
        #endif
        /* switches for the tests */
        else if (strcmp(argv[i]+1, "getpgid") == 0)                 SetOption(argv[i], &opt_getpgid);
        else if (strcmp(argv[i]+1, "getpriority") == 0)             SetOption(argv[i], &opt_getpriority);
        else if (strcmp(argv[i]+1, "getsid") == 0)                  SetOption(argv[i], &opt_getsid);
        else if (strcmp(argv[i]+1, "kill") == 0)                    SetOption(argv[i], &opt_kill);
        #ifdef HAS_SIGQUEUE
        else if (strcmp(argv[i]+1, "sigqueue") == 0)                SetOption(argv[i], &opt_sigqueue);
        #endif
        #ifdef HAS_TKILL
        else if (strcmp(argv[i]+1, "tkill") == 0)                   SetOption(argv[i], &opt_tkill);
        #endif
        #ifdef HAS_SCHEDULER_H
        else if (strcmp(argv[i]+1, "sched_getparam") == 0)          SetOption(argv[i], &opt_sched_getparam);
        else if (strcmp(argv[i]+1, "sched_rr_get_interval") == 0)   SetOption(argv[i], &opt_sched_rr_get_interval);
        else if (strcmp(argv[i]+1, "sched_getscheduler") == 0)      SetOption(argv[i], &opt_sched_getscheduler);
        #ifdef SCHED_GETAFFINITY
        else if (strcmp(argv[i]+1, "sched_getaffinity") == 0)       SetOption(argv[i], &opt_sched_getaffinity);
        #endif
        #endif
        #ifdef HAS_CAPABILITY
        else if (strcmp(argv[i]+1, "capgetp") == 0)                 SetOption(argv[i], &opt_capgetp);
        #endif
        /* tests on the more visible parts of the system */
        else if (strcmp(argv[i]+1, "chdir") == 0)                   SetOption(argv[i], &opt_chdir);
        else if (strcmp(argv[i]+1, "opendir") == 0)                 SetOption(argv[i], &opt_opendir);
        else if (strcmp(argv[i]+1, "proc") == 0)                    SetOption(argv[i], &opt_proc);
        else if (strcmp(argv[i]+1, "readdir") == 0)                 SetOption(argv[i], &opt_readdir);
        #ifdef OS_LINUX
        else if (strcmp(argv[i]+1, "tgid") == 0)                    SetOption(argv[i], &opt_tgid);
        #endif
        /* brute force tests */
        #if defined(HAS_PTHREAD) && defined(OS_LINUX)
        else if (strcmp(argv[i]+1, "pthread") == 0)                 SetOption(argv[i], &opt_pthread);
        #endif
        else if (strcmp(argv[i]+1, "vfork") == 0)                   SetOption(argv[i], &opt_vfork);
        #if defined(OS_LINUX) || defined(OS_SOLARIS)
        /* other tests */
        else if (strcmp(argv[i]+1, "fake_name") == 0)               SetOption(argv[i], &opt_fake_name);
        /* fake search - can be given multiple times */
        else if (strncmp(argv[i], "-f", 2) == 0)
        {
            pc = argv[i];
            while (*++pc == 'f') ++opt_nFakeSearch;
            if (*pc) InvalidOption(pc, INVALID);
        }
        #endif
        /* random pids - can be given multiple times */
        else if (strncmp(argv[i], "-r", 2) == 0)
        {
            pc = argv[i];
            while (*++pc == 'r') ++opt_nRandom;
            if (*pc) InvalidOption(pc, INVALID);
            g_soKernel.bHasRandomPIDs = 1;
        }
        /* verbose - can be given multiple times */
        else if (strncmp(argv[i], "-v", 2) == 0)
        {
            pc = argv[i];
            while (*++pc == 'v') ++opt_nVerbose;
            if (*pc) InvalidOption(pc, INVALID);
        }
        else InvalidOption(argv[i], INVALID);
    }

    if (opt_nListTests)
    {
        PrintTests();
        return 0;
    }

    if (opt_nVerbose > 3) Output(3, "Info: Maximum verbose level is 3 -- no need to increase it behind that.\n");

    g_uUID = getuid();

    #if defined(OS_LINUX) || defined(OS_SOLARIS)
    if (!opt_fake_name && opt_nFakeSearch)
    {
        Output(3, "Activating fake_name via -f option.\n");
        opt_fake_name = 1;
    }
    #endif
    #ifdef OS_SOLARIS
    if (opt_fake_name && g_uUID)
    {
        fprintf(stderr, "Fake process name detection needs root permission to work.\nDisable it with -fake_name\n");
        exit(101);
    }
    #endif

    if (opt_nNiceness)
    {
        if (g_uUID == 0)
        {
            if (setpriority(PRIO_PROCESS, 0, -20))
                Output(0, "Can't set nice value to -20: %s\n", strerror(errno));
        }
        else Output(0, "You have to be root or you can't change process priority.\n");
    }
    if (g_uUID != 0) Output(1, "It's recommended to run %s as root.\n", g_szProgramName);
    return SearchHidden();
}

